"""
CARPI DASH DAEMON
(C) 2018, Raphael "rGunti" Guntersweiler
Licensed under MIT
"""

from carpicommons.log import DEFAULT_CONFIG
from logging import DEBUG

from daemoncommons.daemon import DaemonRunner
from dashdaemon.daemon import DashDaemon

if __name__ == '__main__':
    # configure_logging()
    DEFAULT_CONFIG['root']['level'] = DEBUG
    d = DaemonRunner('DASH_DAEMON_CFG', ['dash.ini', '/etc/carpi/dash.ini'])
    d.run(DashDaemon())



"""
CARPI DASH DAEMON
(C) 2018, Raphael "rGunti" Guntersweiler
Licensed under MIT
"""

from redisdatabus.bus import TypedBusListener as Types
import gpsdaemon.keys as gpskeys
import obddaemon.keys as obdkeys


SETTINGS_KEY_BASE = 'carpi.settings.'
DASH_KEY_BASE = 'carpi.dashboard.'


def _build_key(type, key_base, name):
    return "{}{}{}".format(type if type else "", key_base, name)


CONFIG_KEYS = {
    'engine_vol': _build_key(Types.TYPE_PREFIX_INT, SETTINGS_KEY_BASE, 'car.enginevolume'),
    'vol_efficency': _build_key(Types.TYPE_PREFIX_INT, SETTINGS_KEY_BASE, 'car.efficency'),
    'fuel_density': _build_key(Types.TYPE_PREFIX_INT, SETTINGS_KEY_BASE, 'car.fueldensity')
}

CONFIG_DEFAULT_VALUES = {
    CONFIG_KEYS['engine_vol']: 1000,
    CONFIG_KEYS['vol_efficency']: 85,
    CONFIG_KEYS['fuel_density']: 745
}

LIVE_INPUT_DATA_KEYS = {
    'car_rpm': obdkeys.KEY_RPM,
    'car_map': obdkeys.KEY_INTAKE_PRESSURE,
    'car_tmp': obdkeys.KEY_INTAKE_TEMP,
    'car_spd': obdkeys.KEY_SPEED,
    'gps_spd': gpskeys.KEY_SPEED,
    'gps_acc_lng': gpskeys.KEY_EPX,
    'gps_acc_lat': gpskeys.KEY_EPY,
    'gps_acc_spd': gpskeys.KEY_EPS
}

LIVE_OUTPUT_DATA_KEYS = {
    'speed': _build_key(Types.TYPE_PREFIX_INT, DASH_KEY_BASE, 'speed'),
    'fuel_usage': _build_key(Types.TYPE_PREFIX_FLOAT, DASH_KEY_BASE, 'fuelusage'),
    'fuel_efficiency': _build_key(Types.TYPE_PREFIX_FLOAT, DASH_KEY_BASE, 'fuelefficiency'),
    'fuel_fail_flag': _build_key(Types.TYPE_PREFIX_BOOL, DASH_KEY_BASE, 'fuelfailflag')
}



"""
CARPI REDIS DATA BUS
(C) 2018, Raphael "rGunti" Guntersweiler
Licensed under MIT
"""

from setuptools import setup

with open('README.md', 'r') as f:
    long_description = f.read()

setup(name='carpi-dashdaemon',
      version='0.1.2',
      description='Daemon that processes data from different sources for a digital car dashboard.',
      long_description=long_description,
      url='https://github.com/rGunti/CarPi-DashDaemon',
      keywords='carpi dash daemon',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.6'
      ],
      author='Raphael "rGunti" Guntersweiler',
      author_email='raphael@rgunti.ch',
      license='MIT',
      packages=['dashdaemon'],
      install_requires=[
          'wheel',
          'carpi-commons',
          'carpi-redisdatabus',
          'carpi-daemoncommons',
          'carpi-gpsdaemon',
          'carpi-obddaemon',
      ],
      zip_safe=False,
      include_package_data=True)



"""
CARPI DASH DAEMON
(C) 2018, Raphael "rGunti" Guntersweiler
Licensed under MIT
"""

GAS_CONSTANT: float = 8.3144598  # J/mol.K
AIR_MOL_MASS: float = 28.9644    # g/mol


def _to_kelvin(deg_c: int):
    return deg_c + 273.15


def calculate_fuel_usage(rpm: int,
                         map: int,
                         in_tmp: int,
                         v_E: float,
                         v_H: float,
                         p_F: int) -> float:
    """
    :param rpm: RPM [RPM] (OBD 010C)
    :param map: Manifold Absolute Pressure [kPa] (OBD 010B)
    :param in_tmp: Intake Air Temperature [°C] (OBD 010F)
    :param v_E: Volumetric Efficiency [0-1, %] (default is 85% => 0.85)
    :param v_H: Engine Volume [l] (ccm / 1000)
    :param p_F: Fuel density [g/l] (E10 = 745 g/l, Gas = 720 - 775 g/l)

    :return: Fuel usage in [l/h]
    """
    # Volumetric Efficiency is a value between 0 and 1 (% value)
    assert 0 <= v_E <= 1
    # Engine Volume should be in a "sensible" range (0.5l / 500ccm - 5l / 5000ccm)
    assert 0.5 <= v_H <= 5

    imap = (rpm * map) / _to_kelvin(in_tmp)
    maf = ((imap / 120) * v_E * v_H) / GAS_CONSTANT

    lps = (maf / 14.7) / p_F
    return lps * 3600


def calculate_fuel_efficiency(spd: int, lph: float) -> float:
    """
    :param spd: Vehicle Speed [km/h] (OBD 010D or GPS, used only for efficiency over distance, 0 to disable)
    :param lph: Fuel usage in [l/h]
    :return: Fuel efficiency in [l/100km]
             Returns 0, if speed is reported as 0 km/h (or lower) or fuel usage is 0 l/h (or lower)
    """
    if spd <= 0 or lph <= 0:
        return float(0)
    return (lph / spd) * 100



"""
CARPI DASH DAEMON
(C) 2018, Raphael "rGunti" Guntersweiler
Licensed under MIT
"""
from logging import Logger
from time import sleep

from carpicommons.log import logger
from carpicommons.errors import CarPiExitException
from carpisettings import IniStore, RedisStore, ConfigStore
from daemoncommons.daemon import Daemon
from dashdaemon.keys import LIVE_INPUT_DATA_KEYS, LIVE_OUTPUT_DATA_KEYS, CONFIG_KEYS
from redisdatabus.bus import BusWriter, TypedBusListener

from dashdaemon.work import calculate_fuel_usage, calculate_fuel_efficiency


class DashDaemonError(CarPiExitException):
    DEFAULT_EXIT_CODE = 0xFD00

    REASON_CONFIG_CONNECTION_INVALID = 0xC0
    REASON_CONFIG_WHEREISIT = 0xC1

    def __init__(self, reason=0x0):
        super().__init__(DashDaemonError.DEFAULT_EXIT_CODE + reason)
        self._reason = reason

    @property
    def reason(self):
        return self._reason


class DashDaemon(Daemon):
    def __init__(self):
        super().__init__("Dash Daemon")
        self._log: Logger = None
        self._current_values = dict()
        self._running = False
        self._reader: TypedBusListener = None

    def _build_config_reader(self) -> ConfigStore:
        cfg_type = self._get_config('Config', 'Type', 'Redis')
        if cfg_type.lower() == "ini":
            ini_path = self._get_config('Config', 'Path', None)
            if not ini_path:
                self._log.error("Failed to load configuration because you didn't specify it")
                raise DashDaemonError(DashDaemonError.REASON_CONFIG_WHEREISIT)

            self._log.info("Loading INI configuration from %s ...", ini_path)
            return IniStore(ini_path)
        elif cfg_type.lower() == "redis":
            url = self._get_config('Config', 'URL', None)
            if url:
                self._log.info("Loading Redis configuration from %s ...", url)
                return RedisStore(url=url)
            else:
                host = self._get_config('Config', 'Host', '127.0.0.1')
                port = self._get_config_int('Config', 'Port', 6379)
                db = self._get_config_int('Config', 'DB', 0)

                self._log.info("Loading Redis configuration from %s:%s/%s ...", host, port, db)
                return RedisStore(host=host, port=port, db=db,
                                  password=self._get_config('Config', 'Password', None))
        else:
            self._log.error("Configuration Type %s is unknown or not implemented", cfg_type)
            raise DashDaemonError(DashDaemonError.REASON_CONFIG_CONNECTION_INVALID)

    def _build_bus_writer(self) -> BusWriter:
        self._log.info("Connecting to Data Target Redis instance ...")
        return BusWriter(host=self._get_config('Destination', 'Host', '127.0.0.1'),
                         port=self._get_config_int('Destination', 'Port', 6379),
                         db=self._get_config_int('Destination', 'DB', 0),
                         password=self._get_config('Destination', 'Password', None))

    def _build_bus_reader(self, channels: list) -> TypedBusListener:
        self._log.info("Connecting to Data Source Redis instance ...")
        return TypedBusListener(channels,
                                host=self._get_config('Source', 'Host', '127.0.0.1'),
                                port=self._get_config_int('Source', 'Port', 6379),
                                db=self._get_config_int('Source', 'DB', 0),
                                password=self._get_config('Source', 'Password', None))

    def startup(self):
        self._log = log = logger(self.name)
        log.info("Starting up %s ...", self.name)

        channels = [
            LIVE_INPUT_DATA_KEYS['car_rpm'],
            LIVE_INPUT_DATA_KEYS['car_spd'],
            LIVE_INPUT_DATA_KEYS['car_map'],
            LIVE_INPUT_DATA_KEYS['car_tmp']
        ]

        config = self._build_config_reader()
        reader = self._build_bus_reader(channels)
        writer = self._build_bus_writer()

        reader.register_global_callback(self._on_new_value_registered)

        reader.start()
        self._running = True

        log.info("Ready to enter main loop")
        while self._running:
            self._step(config, writer)
            sleep(0.2)

        self._reader.stop()

    def shutdown(self):
        self._running = False
        if self._reader:
            self._reader.stop()

    def _on_new_value_registered(self, channel, value):
        self._log.debug("Reported new value from %s: %s", channel, value)
        self._current_values[channel] = value

    def _step(self, config: ConfigStore, writer: BusWriter):
        values = self._current_values

        # Pass over speed directly
        # TODO: In the future, compare GPS and OBD speed and try to find the best match
        car_speed = values.get(LIVE_INPUT_DATA_KEYS['car_spd'], 0)
        writer.publish(LIVE_OUTPUT_DATA_KEYS['speed'], car_speed)

        try:
            # Calculate Fuel Efficiency
            rpm = values[LIVE_INPUT_DATA_KEYS['car_rpm']]
            map = values[LIVE_INPUT_DATA_KEYS['car_map']]
            in_tmp = values[LIVE_INPUT_DATA_KEYS['car_tmp']]

            vol_efficiency = config.read_int_value(CONFIG_KEYS['vol_efficency'], 85) / 100
            eng_volume = config.read_int_value(CONFIG_KEYS['engine_vol']) / 1000
            fuel_density = config.read_int_value(CONFIG_KEYS['fuel_density'], 745)

            lph = calculate_fuel_usage(rpm, map, in_tmp,
                                       vol_efficiency, eng_volume, fuel_density)
            lpk = calculate_fuel_efficiency(car_speed, lph)

            writer.publish(LIVE_OUTPUT_DATA_KEYS['fuel_usage'], lph)
            writer.publish(LIVE_OUTPUT_DATA_KEYS['fuel_efficiency'], lpk)
        except KeyError as e:
            # If values aren't present, just fail fast
            writer.publish(LIVE_OUTPUT_DATA_KEYS['fuel_fail_flag'], True)
